# GPIO in Linux

## GPIO Kernel API
\subsection{Kernel API}

* Legacy integer-based API  `[gpio]`  
`linux/gpio.h`
* Current descriptor-based API `[gpiod]`  
`linux/gpio/consumer.h`
* GPIO should be obtained (reserved) before used
* GPIO can be exported to userspace

## Kernel example

```C
// From Documentation/gpio/board.txt
foo_device {
		compatible = "acme,foo";
		...
		led-gpios = <&gpio 15 GPIO_ACTIVE_HIGH>, /* red */
			    <&gpio 16 GPIO_ACTIVE_HIGH>, /* green */
			    <&gpio 17 GPIO_ACTIVE_HIGH>; /* blue */

		power-gpios = <&gpio 1 GPIO_ACTIVE_LOW>;
	};
```

## Kernel example

```C
// From Documentation/gpio/board.txt
struct gpio_desc *red, *green, *blue, *power;

red = gpiod_get_index(dev, "led", 0, GPIOD_OUT_HIGH);
green = gpiod_get_index(dev, "led", 1, GPIOD_OUT_HIGH);
blue = gpiod_get_index(dev, "led", 2, GPIOD_OUT_HIGH);
power = gpiod_get(dev, "power", GPIOD_OUT_HIGH);
```

## Access functions

```C
int  gpiod_direction_input(struct gpio_desc *desc)
int  gpiod_direction_output(struct gpio_desc *desc, int value)
```

* Two variants of set/get functions:
  * Functions that are spinlock-safe (controller is memory mapped)
  * Functions that can sleep (controller connected via external bus - I2C, etc.)

```C
int  gpiod_get_value(const sruct gpio_desc *desc);
void gpiod_set_value(struct gpio_desc *desc, int value);
int  gpiod_get_value_cansleep(const struct gpio_desc *desc);
void gpiod_set_value_cansleep(struct gpio_desc *desc, int value);
```

\subsection{Sysfs API}
## Userspace - Sysfs API

* Legacy userspace API in `/sys/class/gpio`
* Currently deprecated
* Remains exported when application crashes
* Multiple file descriptors, multiple syscalls

```bash
# Export GPIO pin 15
echo 15 > /sys/class/gpio/export
# Set as output
echo out > /sys/class/gpio/gpio15/direction
# Set GPIO 15 to "1"
echo 1 > /sys/class/gpio/gpio15/value
```

\subsection{Character device API}
## Userspace - Character device API

* Merged in 4.8
* One device per gpiochip
* Access via `/dev/gpiochip0` etc.
* Allows multiple operations at single syscall
* API defined in `include/linux/gpio.h`
* **libgpiod** - C library for handling new GPIO userspace API
* Userspace tools for GPIO handling provided
* <https://git.kernel.org/pub/scm/libs/libgpiod/libgpiod.git>

## libgpiod - example

```C
	struct gpiod_chip *chip;
	struct gpiod_line *line;
	// Open GPIO chip
	chip = gpiod_chip_open("/dev/gpiochip0");
	// Get line (pin) from the GPIO chip
	line = gpiod_chip_get_line(chip, offset);
	// Request (reserve) the line and set as output
	gpiod_line_request_output(line, "consumer", 0);
	// Set value to the line
	gpiod_line_set_value(line, 0);
	// Release the line
	gpiod_line_release(line);
```

## libgpiod - Tools

* **gpiodetect** - list all gpiochips present on the system, their names, labels
               and number of GPIO lines
* **gpioinfo**   - list all lines of specified gpiochips, their names, consumers,
               direction, active state and additional flags
* **gpioget**    - read values of specified GPIO lines
* **gpioset**    - set values of specified GPIO lines, potentially keep the lines
               exported and wait until timeout, user input or signal
* **gpiofind**   - find the gpiochip name and line offset given the line name
* **gpiomon**    - wait for events on GPIO lines, specify which events to watch,
               how many events to process before exiting or if the events
               should be reported to the console

\subsection{Library}
## Library

[GPIO in the kernel: an introduction](https://lwn.net/Articles/532714/)  
[GPIO Sysfs Interface for Userspace](https://www.kernel.org/doc/html/latest/admin-guide/gpio/sysfs.html)  
[New GPIO interface for Userspace](https://schd.ws/hosted_files/osseu17/88/elce2017_new_GPIO_interface.pdf)  
[Documentation/gpio/board.txt]()
