/* __init and __exit macros */
#include <linux/init.h>
/* module_init, module_exit, meta information macros */
#include <linux/module.h>
/* Kernel device model */
#include <linux/device.h>
/* Common kernel macros. */
#include <linux/kernel.h>
/* struct file_operations, chardev */
#include <linux/fs.h>
/* Userspace access - copy from user, copy to user ... */
#include <linux/uaccess.h>

/* This name will appear at /dev/ */
#define  DEVICE_NAME "fcdev"
/* Device class - will appear in /sys/class */
#define  CLASS_NAME  "ourclass"
/* Maximal length of message written to the device */
#define  MAX_themsg_LEN 1024

/* Major device number */
static int    major;
/* Storage for the message */
static char   themsg[MAX_themsg_LEN] = {0};
/* Current length of the message */
static short  themsg_len;
/* How many times the device was opened */
static int    open_cnt = 0;
static struct class*  fcdev_class  = NULL;
static struct device* fcdev_dev = NULL;

/* Declarations of file_operations functions */
static int     fcdev_open(struct inode *, struct file *);
static int     fcdev_release(struct inode *, struct file *);
static ssize_t fcdev_read(struct file *, char *, size_t, loff_t *);
static ssize_t fcdev_write(struct file *, const char *, size_t, loff_t *);

/* Assign our functions to file_operations assigned to the device */
static struct file_operations fcdev_fops = {
    .open = fcdev_open,
    .read = fcdev_read,
    .write = fcdev_write,
    .release = fcdev_release,
};

/* Module entrypoint */
static int __init fcdev_init(void)
{
    printk(KERN_INFO "fcdev: Module init\n");

    /* Dynamically allocate a major number for the dev */
    major = register_chrdev(0, DEVICE_NAME, &fcdev_fops);
    if (major < 0) {
        printk(KERN_ALERT "fcdev: Failed to register a major number\n");
        return 1;
    }

    printk(KERN_INFO "fcdev: Chardev registered with major: %d\n", major);

    /* Register the class */
    fcdev_class = class_create(THIS_MODULE, CLASS_NAME);
    if (IS_ERR(fcdev_class)) {
        unregister_chrdev(major, DEVICE_NAME);
        printk(KERN_ALERT "Failed to register device class\n");

        return PTR_ERR(fcdev_class);
    }
    printk(KERN_INFO "fcdev: device class registered!\n");

    /* Register the device driver */
    fcdev_dev = device_create(fcdev_class, NULL, MKDEV(major, 0), NULL, DEVICE_NAME);
    if (IS_ERR(fcdev_dev)) {
        class_destroy(fcdev_class);
        unregister_chrdev(major, DEVICE_NAME);
        printk(KERN_ALERT "Failed to create the device\n");

        return PTR_ERR(fcdev_dev);
    }

    printk(KERN_INFO "fcdev: device class created!\n");
    return 0;
}

/* Module exit point. */
static void __exit fcdev_exit(void)
{
    device_destroy(fcdev_class, MKDEV(major, 0));
    class_unregister(fcdev_class);
    class_destroy(fcdev_class);
    unregister_chrdev(major, DEVICE_NAME);
    printk(KERN_INFO "fcdev: Unloaded!\n");
}

/* Implementation of OPEN for file_operations. Called when user calls open() */
static int fcdev_open(struct inode *inodep, struct file *filep)
{
    open_cnt++;
    printk(KERN_INFO "fcdev: Device has been opened %d time(s)\n", open_cnt);

    return 0;
}

/* Implementation of READ for file_operations. Driver -> User */
static ssize_t fcdev_read(struct file *filep, char *buffer, size_t len, loff_t *offset)
{
    int err = 0;
    ssize_t written = themsg_len;

    /* Copy buffer to userspace */
    err = copy_to_user(buffer, themsg, themsg_len);

    if (err == 0) {
        printk(KERN_INFO "fcdev: Sent %zd characters to the user\n", written);
        themsg_len = 0;

        return written;
    } else {
        printk(KERN_INFO "fcdev: Failed to send characters to the user [%d]\n", err);

        return -EFAULT;
    }
}

/* Implementation of WRITE for file_operations. User -> Driver */
static ssize_t fcdev_write(struct file *filep, const char *buffer, size_t len, loff_t *offset)
{
    copy_from_user(themsg, buffer, len);
    themsg_len = len;
    printk(KERN_INFO "fcdev: Received %zu bytes from the user [%d chars] (%s)\n", len, themsg_len, themsg);
    return len;
}

/* Implementation of RELEASE for file_operations. Called when user calls close() */
static int fcdev_release(struct inode *inodep, struct file *filep)
{
    printk(KERN_INFO "fcdev: Device successfully closed\n");
    return 0;
}

module_init(fcdev_init);
module_exit(fcdev_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Acrios");
MODULE_DESCRIPTION("Our first character driver");
MODULE_VERSION("1.0");
